from datasette import Response, Forbidden, NotFound
from datasette_acl.utils import (
    can_edit_permissions,
    get_acl_actor_ids,
    validate_actor_id,
)
import json
import re

GROUPS_SQL = """
select
    acl_groups.id,
    acl_groups.name,
    acl_groups.deleted,
    count(acl_actor_groups.actor_id) as size,
    json_group_array(
        acl_actor_groups.actor_id
    ) filter (
        where
        acl_actor_groups.actor_id is not null
    ) as actor_ids
from
    acl_groups
left join
    acl_actor_groups on acl_groups.id = acl_actor_groups.group_id
{extra_where}
group by
    acl_groups.id, acl_groups.name
order by
    acl_groups.name;
"""


def get_dynamic_groups(datasette):
    config = datasette.plugin_config("datasette-acl")
    return config.get("dynamic-groups") or {}


_group_name_re = re.compile(r"^[a-zA-Z0-9_-]+$")


def is_valid_group_name(new_group):
    return bool(_group_name_re.match(new_group))


async def manage_groups(request, datasette):
    if not await can_edit_permissions(datasette, request.actor):
        raise Forbidden("You do not have permission to edit permissions")
    internal_db = datasette.get_internal_database()
    groups = [
        dict(r, actor_ids=json.loads(r["actor_ids"]))
        for r in await internal_db.execute(
            GROUPS_SQL.format(extra_where=" where deleted is null")
        )
    ]
    if request.method == "POST":
        post_vars = await request.post_vars()
        new_group = (post_vars.get("new_group") or "").strip()
        if new_group:
            # Is it valid?
            if (
                await internal_db.execute(
                    "select 1 from acl_groups where name = :name and deleted is null",
                    {"name": new_group},
                )
            ).first():
                datasette.add_message(
                    request, "This group already exists", datasette.ERROR
                )
                return Response.redirect(
                    datasette.urls.path("/-/acl/groups/" + new_group)
                )
            elif not is_valid_group_name(new_group):
                datasette.add_message(
                    request,
                    "Group names must use characters a-zA-Z0-0_-",
                    datasette.ERROR,
                )
                return Response.redirect(datasette.urls.path("/-/acl/groups"))
            else:
                # Create group if it does not exist
                await internal_db.execute_write(
                    "insert or ignore into acl_groups (name) values (:name)",
                    {"name": new_group},
                )
                # Ensure it is not marked as deleted
                await internal_db.execute_write(
                    "update acl_groups set deleted = null where name = :name",
                    {"name": new_group},
                )
                # Audit log record
                await internal_db.execute_write(
                    """
                    insert into acl_groups_audit (
                        operation_by, operation, group_id
                    ) values (
                        :operation_by,
                        'created',
                        (select id from acl_groups where name = :group_name)
                    )
                """,
                    {"operation_by": request.actor["id"], "group_name": new_group},
                )
                datasette.add_message(request, f"Group created: {new_group}")
                return Response.redirect(
                    datasette.urls.path("/-/acl/groups/" + new_group)
                )

    dynamic_groups = get_dynamic_groups(datasette)
    return Response.html(
        await datasette.render_template(
            "manage_acl_groups.html",
            {
                "groups": groups,
                "dynamic_groups": dynamic_groups,
            },
            request=request,
        )
    )


async def manage_group(request, datasette):
    if not await can_edit_permissions(datasette, request.actor):
        raise Forbidden("You do not have permission to edit permissions")
    name = request.url_vars["name"]
    internal_db = datasette.get_internal_database()
    group = (
        await internal_db.execute(
            GROUPS_SQL.format(extra_where=" where acl_groups.name = :name"),
            {
                "name": name,
            },
        )
    ).first()
    if not group:
        raise NotFound("Group does not exist")
    group_id = group["id"]
    dynamic_groups = get_dynamic_groups(datasette)
    dynamic_config = dynamic_groups.get(name)
    actor_ids = json.loads(group["actor_ids"])

    async def audit_log(*, operation, actor_id=None):
        await internal_db.execute_write(
            f"""
            insert into acl_groups_audit (
                operation_by, operation, group_id, actor_id
            ) values (
                :operation_by,
                :operation,
                :group_id,
                {':actor_id' if actor_id else 'null'}
            )
        """,
            {
                "operation_by": request.actor["id"],
                "operation": operation,
                "group_id": group_id,
                "actor_id": actor_id,
            },
        )

    async def remove_member(actor_id):
        await internal_db.execute_write(
            """
            delete from acl_actor_groups
            where actor_id = :actor_id
            and group_id = :group_id
        """,
            {"actor_id": actor_id, "group_id": group_id},
        )
        await audit_log(operation="removed", actor_id=actor_id)

    if request.method == "POST" and not dynamic_config:
        post_vars = await request.post_vars()
        to_add = post_vars.get("add")
        to_remove = post_vars.get("remove")

        should_delete = post_vars.get("delete_group")
        if should_delete:
            # First remove all the members
            for actor_id in actor_ids:
                await remove_member(actor_id)
            # Now mark the group as deleted and record
            await internal_db.execute_write(
                "update acl_groups set deleted = 1 where id = :group_id",
                {"group_id": group_id},
            )
            await audit_log(operation="deleted")
            datasette.add_message(request, f"Group deleted: {name}")
            return Response.redirect(datasette.urls.path("/-/acl/groups"))

        fragment = ""
        if to_remove:
            if to_remove not in actor_ids:
                datasette.add_message(
                    request, "That user is not in the group", datasette.ERROR
                )
            else:
                await remove_member(to_remove)
                datasette.add_message(request, f"Removed {to_remove}")
        if to_add:
            if to_add in actor_ids:
                datasette.add_message(
                    request, "That user is already in the group", datasette.ERROR
                )
            else:
                # Add user
                if not await validate_actor_id(datasette, to_add):
                    datasette.add_message(
                        request, "That user ID is not valid", datasette.ERROR
                    )
                    return Response.redirect(request.path)
                await internal_db.execute_write(
                    """
                    insert into acl_actor_groups (actor_id, group_id)
                    values (:actor_id, :group_id)
                """,
                    {"actor_id": to_add, "group_id": group_id},
                )
                datasette.add_message(request, f"Added {to_add}")
                await audit_log(operation="added", actor_id=to_add)
                fragment = "#focus-add"
        return Response.redirect(request.path + fragment)

    return Response.html(
        await datasette.render_template(
            "manage_acl_group.html",
            {
                "name": name,
                "size": group["size"],
                "is_deleted": group["deleted"],
                "members": actor_ids,
                "dynamic_config": dynamic_config,
                "audit_log": [
                    dict(r)
                    for r in await internal_db.execute(
                        """
                        select
                            timestamp, operation_by, operation, actor_id
                        from acl_groups_audit
                        where group_id = ?
                        order by id desc
                    """,
                        [group_id],
                    )
                ],
                "valid_actor_ids": await get_acl_actor_ids(datasette),
            },
            request=request,
        )
    )



from collections import namedtuple
from datasette.app import Datasette
from datasette_acl import update_dynamic_groups
import pytest


ManageTableTest = namedtuple(
    "ManageTableTest",
    (
        "description",
        "setup_post_data",
        "post_data",
        "expected_acls",
        "should_fail_then_succeed",
        "expected_audit_rows",
    ),
)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ManageTableTest._fields,
    (
        ManageTableTest(
            description="Group: add insert-row",
            setup_post_data={},
            post_data={"group_permissions_staff_insert-row": "on"},
            expected_acls=[
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                }
            ],
            should_fail_then_succeed=[
                dict(
                    actor={"id": "simon", "is_staff": True},
                    action="insert-row",
                    resource=["db", "t"],
                ),
            ],
            expected_audit_rows=[
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                }
            ],
        ),
        ManageTableTest(
            description="Group: remove insert-row, add update-row and delete-row",
            setup_post_data={"group_permissions_staff_insert-row": "on"},
            post_data={
                "group_permissions_staff_update-row": "on",
                "group_permissions_staff_delete-row": "on",
            },
            expected_acls=[
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "delete-row",
                    "database_name": "db",
                    "resource_name": "t",
                },
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "update-row",
                    "database_name": "db",
                    "resource_name": "t",
                },
            ],
            should_fail_then_succeed=[
                dict(
                    actor={"id": "simon", "is_staff": True},
                    action="delete-row",
                    resource=["db", "t"],
                ),
                dict(
                    actor={"id": "simon", "is_staff": True},
                    action="update-row",
                    resource=["db", "t"],
                ),
            ],
            expected_audit_rows=[
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "removed",
                },
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "delete-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
                {
                    "group_name": "staff",
                    "actor_id": None,
                    "action_name": "update-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
            ],
        ),
        ManageTableTest(
            description="New user: set with insert-row and update-row",
            setup_post_data={},
            post_data={
                "new_actor_id": "newbie",
                "new_user_insert-row": "on",
                "new_user_update-row": "on",
            },
            expected_acls=[
                {
                    "action_name": "insert-row",
                    "actor_id": "newbie",
                    "database_name": "db",
                    "group_name": None,
                    "resource_name": "t",
                },
                {
                    "action_name": "update-row",
                    "actor_id": "newbie",
                    "database_name": "db",
                    "group_name": None,
                    "resource_name": "t",
                },
            ],
            should_fail_then_succeed=[
                dict(
                    actor={"id": "newbie"},
                    action="insert-row",
                    resource=["db", "t"],
                ),
                dict(
                    actor={"id": "newbie"},
                    action="update-row",
                    resource=["db", "t"],
                ),
            ],
            expected_audit_rows=[
                {
                    "group_name": None,
                    "actor_id": "newbie",
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
                {
                    "group_name": None,
                    "actor_id": "newbie",
                    "action_name": "update-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
            ],
        ),
        ManageTableTest(
            description="Existing user: remove insert-row, add update-row",
            setup_post_data={
                "new_actor_id": "newbie",
                "new_user_insert-row": "on",
            },
            post_data={
                "user_permissions_newbie_update-row": "on",
            },
            expected_acls=[
                {
                    "action_name": "update-row",
                    "actor_id": "newbie",
                    "database_name": "db",
                    "group_name": None,
                    "resource_name": "t",
                }
            ],
            should_fail_then_succeed=[
                dict(
                    actor={"id": "newbie"},
                    action="update-row",
                    resource=["db", "t"],
                ),
            ],
            expected_audit_rows=[
                {
                    "group_name": None,
                    "actor_id": "newbie",
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
                {
                    "group_name": None,
                    "actor_id": "newbie",
                    "action_name": "insert-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "removed",
                },
                {
                    "group_name": None,
                    "actor_id": "newbie",
                    "action_name": "update-row",
                    "database_name": "db",
                    "resource_name": "t",
                    "operation_by": "root",
                    "operation": "added",
                },
            ],
        ),
    ),
)
async def test_manage_table_permissions(
    ds,
    csrftoken,
    description,
    setup_post_data,
    post_data,
    expected_acls,
    should_fail_then_succeed,
    expected_audit_rows,
):
    internal_db = ds.get_internal_database()

    # Staff dynamic group should have been created on startup
    assert (
        await internal_db.execute(
            "select count(*) from acl_groups where name = 'staff'"
        )
    ).single_value() == 1

    if setup_post_data:
        setup_response = await ds.client.post(
            "/db/t/-/acl",
            data={**setup_post_data, "csrftoken": csrftoken},
            cookies={
                "ds_actor": ds.client.actor_cookie({"id": "root"}),
                "ds_csrftoken": csrftoken,
            },
        )
        assert setup_response.status_code == 302

    # Permission checks should fail
    for kwargs in should_fail_then_succeed:
        assert not await ds.permission_allowed(**kwargs), f"Should have failed: {repr}"

    # Use the /db/table/-/acl page to update permissions
    response = await ds.client.post(
        "/db/t/-/acl",
        data={**post_data, "csrftoken": csrftoken},
        cookies={
            "ds_actor": ds.client.actor_cookie({"id": "root"}),
            "ds_csrftoken": csrftoken,
        },
    )
    assert response.status_code == 302

    # Check ACLs
    acls = [
        dict(r)
        for r in (
            await internal_db.execute(
                """
        select
          acl_groups.name as group_name,
          acl.actor_id,
          acl_actions.name as action_name,
          acl_resources.database as database_name,
          acl_resources.resource as resource_name
        from acl
        left join acl_groups on acl.group_id = acl_groups.id
        join acl_actions on acl.action_id = acl_actions.id
        join acl_resources on acl.resource_id = acl_resources.id
    """
            )
        )
    ]
    assert acls == expected_acls

    # Permission checks should pass now
    for kwargs in should_fail_then_succeed:
        assert await ds.permission_allowed(
            **kwargs
        ), f"Should have passed: {repr(kwargs)}"

    # Check audit logs
    AUDIT_SQL = """
        select
          acl_groups.name as group_name,
          acl_audit.actor_id,
          acl_actions.name as action_name,
          acl_resources.database as database_name,
          acl_resources.resource as resource_name,
          acl_audit.operation_by,
          acl_audit.operation
        from acl_audit
        left join acl_groups on acl_audit.group_id = acl_groups.id
        join acl_actions on acl_audit.action_id = acl_actions.id
        join acl_resources on acl_audit.resource_id = acl_resources.id
        order by acl_audit.id
    """
    audit_rows = [dict(r) for r in (await internal_db.execute(AUDIT_SQL))]
    assert audit_rows == expected_audit_rows


@pytest.mark.asyncio
async def test_update_dynamic_groups():
    datasette = Datasette(
        config={
            "plugins": {
                "datasette-acl": {
                    "dynamic-groups": {
                        "staff": {"is_staff": True},
                    }
                }
            }
        }
    )
    await datasette.invoke_startup()
    db = datasette.get_internal_database()
    # Should have those tables
    tables = await db.table_names()
    assert {
        "acl_actions",
        "acl_actor_groups",
        "acl_audit",
        "acl_groups_audit",
        "acl_groups",
        "acl_resources",
        "acl",
    }.issubset(tables)
    # Group tables should start populated
    assert (await db.execute("select count(*) from acl_groups")).single_value() == 1
    # But no actor groups
    assert (
        await db.execute("select count(*) from acl_actor_groups")
    ).single_value() == 0
    # An actor with is_staff: True should be added to the group
    await update_dynamic_groups(
        datasette, {"is_staff": True, "id": "staff"}, skip_cache=True
    )
    assert [dict(r) for r in (await db.execute("select * from acl_groups")).rows] == [
        {"id": 1, "name": "staff", "deleted": None},
    ]
    # Should record an added groups audit record
    assert (
        [
            dict(r)
            for r in (
                await db.execute(
                    """
            select operation, operation_by, group_id, actor_id
            from acl_groups_audit
        """
                )
            ).rows
        ]
        == [
            {
                "operation": "added",
                "operation_by": None,
                "group_id": 1,
                "actor_id": "staff",
            },
        ]
    )
    assert [
        dict(r)
        for r in (
            await db.execute(
                "select actor_id, (select name from acl_groups where id = group_id) as group_name from acl_actor_groups"
            )
        ).rows
    ] == [
        {"actor_id": "staff", "group_name": "staff"},
    ]
    # If that user changes they should drop from the group
    await update_dynamic_groups(
        datasette, {"is_staff": False, "id": "staff"}, skip_cache=True
    )
    assert [
        dict(r)
        for r in (
            await db.execute(
                "select actor_id, (select name from acl_groups where id = group_id) as group_name from acl_actor_groups"
            )
        ).rows
    ] == []
    # Should record a removed groups audit record
    assert (
        [
            dict(r)
            for r in (
                await db.execute(
                    """
            select operation, operation_by, group_id, actor_id
            from acl_groups_audit order by id desc limit 1
        """
                )
            ).rows
        ]
        == [
            {
                "operation": "removed",
                "operation_by": None,
                "group_id": 1,
                "actor_id": "staff",
            },
        ]
    )
    # Groups that are not dynamic should not be modified
    await db.execute_write("insert into acl_groups (id, name) values (2, 'static')")
    await db.execute_write(
        "insert into acl_actor_groups (actor_id, group_id) values ('staff', 2)"
    )
    await update_dynamic_groups(
        datasette, {"is_staff": False, "id": "staff"}, skip_cache=True
    )
    assert [dict(r) for r in (await db.execute("select * from acl_groups")).rows] == [
        {"id": 1, "name": "staff", "deleted": None},
        {"id": 2, "name": "static", "deleted": None},
    ]


@pytest.mark.asyncio
async def test_table_creator_permissions():
    datasette = Datasette(
        config={
            "plugins": {
                "datasette-acl": {
                    "table-creator-permissions": [
                        "insert-row",
                        "delete-row",
                    ]
                }
            },
            "permissions": {"create-table": {"id": "*"}},
        }
    )
    await datasette.invoke_startup()
    datasette.add_memory_database("db")
    # Create a table
    actor_cookie = datasette.client.actor_cookie({"id": "simon"})
    create_response = await datasette.client.post(
        "/db/-/create",
        json={
            "table": "new_table",
            "columns": [
                {"name": "id", "type": "integer"},
                {"name": "title", "type": "text"},
            ],
            "pk": "id",
        },
        cookies={"ds_actor": actor_cookie},
    )
    assert create_response.status_code == 201
    # That table should have insert-row and delete-row ACLs
    acls = [
        dict(r)
        for r in (
            await datasette.get_internal_database().execute(
                """
        select
          acl.actor_id,
          acl_actions.name as action_name,
          acl_resources.database as database_name,
          acl_resources.resource as resource_name
        from acl
        join acl_actions on acl.action_id = acl_actions.id
        join acl_resources on acl.resource_id = acl_resources.id
        where acl_resources.database = 'db'
        and acl_resources.resource = 'new_table'
    """
            )
        )
    ]
    assert acls == [
        {
            "actor_id": "simon",
            "action_name": "insert-row",
            "database_name": "db",
            "resource_name": "new_table",
        },
        {
            "actor_id": "simon",
            "action_name": "delete-row",
            "database_name": "db",
            "resource_name": "new_table",
        },
    ]
    # Permission checks too
    assert await datasette.permission_allowed(
        actor={"id": "simon"}, action="insert-row", resource=["db", "new_table"]
    )
    assert await datasette.permission_allowed(
        actor={"id": "simon"}, action="delete-row", resource=["db", "new_table"]
    )
    assert not await datasette.permission_allowed(
        actor={"id": "simon"}, action="update-row", resource=["db", "new_table"]
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("should_work", (True, False))
async def test_table_actions(ds, should_work):
    response = await ds.client.get(
        "/db/t",
        cookies={
            "ds_actor": ds.client.actor_cookie(
                {"id": "root" if should_work else "other"}
            ),
        },
    )
    fragment = '<a href="/db/t/-/acl">Manage table permissions'
    if should_work:
        assert fragment in response.text
    else:
        assert fragment not in response.text



from datasette import Response, Forbidden
from datasette_acl.utils import (
    can_edit_permissions,
    generate_changes_message,
    get_acl_actor_ids,
    validate_actor_id,
)


async def manage_table_acls(request, datasette):
    if not await can_edit_permissions(datasette, request.actor):
        raise Forbidden("You do not have permission to edit permissions")
    table = request.url_vars["table"]
    database = request.url_vars["database"]
    internal_db = datasette.get_internal_database()
    groups = [
        g["name"]
        for g in await datasette.get_internal_database().execute(
            "select name from acl_groups"
        )
    ]

    # Ensure we have a resource_id for this table
    await internal_db.execute_write(
        "INSERT OR IGNORE INTO acl_resources (database, resource) VALUES (?, ?);",
        [database, table],
    )
    resource_id = (
        await internal_db.execute(
            "SELECT id FROM acl_resources WHERE database = ? AND resource = ?",
            [database, table],
        )
    ).single_value()

    current_group_permissions = {}
    current_user_permissions = {}
    acl_rows = await internal_db.execute(
        """
        select
          acl_groups.name as group_name,
          acl.actor_id,
          acl_actions.name as action_name
        from acl
        left join acl_groups on acl.group_id = acl_groups.id
        join acl_actions on acl.action_id = acl_actions.id
        where acl.resource_id = ?
        """,
        [resource_id],
    )
    for row in acl_rows.rows:
        group_name = row["group_name"]
        actor_id = row["actor_id"]
        action_name = row["action_name"]
        if group_name:
            current_group_permissions.setdefault(group_name, {})[action_name] = True
            current_group_permissions[group_name][action_name] = True
        else:
            assert actor_id
            current_user_permissions.setdefault(actor_id, {})[action_name] = True

    if request.method == "POST":
        group_changes_made = {"added": [], "removed": []}
        post_vars = await request.post_vars()
        for group_name in groups:
            for action_name in [
                "insert-row",
                "delete-row",
                "update-row",
                "alter-table",
                "drop-table",
            ]:
                new_value = bool(
                    post_vars.get(f"group_permissions_{group_name}_{action_name}")
                )
                current_value = bool(
                    current_group_permissions.get(group_name, {}).get(action_name)
                )
                if new_value != current_value:
                    if new_value:
                        # They added it, add the record
                        await internal_db.execute_write(
                            """
                            INSERT INTO acl (actor_id, group_id, resource_id, action_id)
                            VALUES (
                                null,
                                (SELECT id FROM acl_groups WHERE name = :group_name),
                                :resource_id,
                                (SELECT id FROM acl_actions WHERE name = :action_name)
                            )
                            """,
                            {
                                "group_name": group_name,
                                "action_name": action_name,
                                "resource_id": resource_id,
                            },
                        )
                        operation = "added"
                        group_changes_made["added"].append((group_name, action_name))
                    else:
                        # They removed it
                        await internal_db.execute_write(
                            """
                            delete from acl where
                                actor_id is null and 
                                group_id = (SELECT id FROM acl_groups WHERE name = :group_name)
                                and resource_id = :resource_id
                                and action_id = (SELECT id FROM acl_actions WHERE name = :action_name)
                            """,
                            {
                                "group_name": group_name,
                                "action_name": action_name,
                                "resource_id": resource_id,
                            },
                        )
                        operation = "removed"
                        group_changes_made["removed"].append((group_name, action_name))
                    await internal_db.execute_write(
                        """
                        insert into acl_audit (
                            operation,
                            actor_id,
                            group_id,
                            resource_id,
                            action_id,
                            operation_by
                        ) values (
                            :operation,
                            null,
                            (SELECT id FROM acl_groups WHERE name = :group_name),
                            :resource_id,
                            (SELECT id FROM acl_actions WHERE name = :action_name),
                            :operation_by
                        )
                        """,
                        {
                            "operation": operation,
                            "group_name": group_name,
                            "resource_id": resource_id,
                            "action_name": action_name,
                            "operation_by": request.actor["id"],
                        },
                    )
        user_changes_made = {"added": [], "removed": []}
        for actor_id in list(current_user_permissions) + [None]:
            if actor_id is None:
                # This is the special case for new_user_{{ action }}
                actor_id = (post_vars.get("new_actor_id") or "").strip()
                if not actor_id:
                    continue
                if not await validate_actor_id(datasette, actor_id):
                    datasette.add_message(
                        request, "That user ID is not valid", datasette.ERROR
                    )
                    return Response.redirect(request.path)
                post_key_prefix = "new_user"
            else:
                post_key_prefix = f"user_permissions_{actor_id}"

            for action_name in [
                "insert-row",
                "delete-row",
                "update-row",
                "alter-table",
                "drop-table",
            ]:
                new_value = bool(post_vars.get(f"{post_key_prefix}_{action_name}"))
                current_value = bool(
                    current_user_permissions.get(actor_id, {}).get(action_name)
                )
                if new_value != current_value:
                    if new_value:
                        # They added the permission
                        await internal_db.execute_write(
                            """
                            insert into acl (actor_id, group_id, resource_id, action_id)
                            values (
                                :actor_id,
                                null,
                                :resource_id,
                                (select id from acl_actions where name = :action_name)
                            )
                            """,
                            {
                                "actor_id": actor_id,
                                "action_name": action_name,
                                "resource_id": resource_id,
                            },
                        )
                        operation = "added"
                        user_changes_made["added"].append((actor_id, action_name))
                    else:
                        # They removed the permission
                        await internal_db.execute_write(
                            """
                            delete from acl where
                                actor_id = :actor_id
                                and group_id is null
                                and resource_id = :resource_id
                                and action_id = (select id from acl_actions where name = :action_name)
                            """,
                            {
                                "actor_id": actor_id,
                                "action_name": action_name,
                                "resource_id": resource_id,
                            },
                        )
                        operation = "removed"
                        user_changes_made["removed"].append((actor_id, action_name))
                    await internal_db.execute_write(
                        """
                        insert into acl_audit (
                            operation,
                            actor_id,
                            group_id,
                            resource_id,
                            action_id,
                            operation_by
                        ) values (
                            :operation,
                            :actor_id,
                            null,
                            :resource_id,
                            (SELECT id FROM acl_actions WHERE name = :action_name),
                            :operation_by
                        )
                        """,
                        {
                            "operation": operation,
                            "actor_id": actor_id,
                            "resource_id": resource_id,
                            "action_name": action_name,
                            "operation_by": request.actor["id"],
                        },
                    )

        if group_changes_made or user_changes_made:
            group_message = generate_changes_message(group_changes_made, "group")
            if group_message:
                datasette.add_message(request, group_message)
            user_message = generate_changes_message(user_changes_made, "user")
            if user_message:
                datasette.add_message(request, user_message)

        return Response.redirect(request.path)

    audit_log = await internal_db.execute(
        """
        select
            acl_audit.timestamp,
            acl_audit.operation_by,
            acl_audit.operation,
            acl_audit.actor_id,
            acl_groups.name as group_name,
            acl_actions.name as action_name
        from acl_audit
        left join acl_groups on acl_audit.group_id = acl_groups.id
        join acl_actions on acl_audit.action_id = acl_actions.id
        where acl_audit.resource_id = ?
        order by acl_audit.timestamp desc
        limit 50
        """,
        [resource_id],
    )

    # group_sizes dictionary for displaying their sizes
    group_sizes = {
        row["name"]: row["size"]
        for row in await internal_db.execute(
            """
            select
                acl_groups.name as name,
                count(acl_actor_groups.actor_id) as size
            from
                acl_groups
            left join
                acl_actor_groups on acl_groups.id = acl_actor_groups.group_id
            group by
                acl_groups.id, acl_groups.name
            """
        )
    }

    return Response.html(
        await datasette.render_template(
            "manage_table_acls.html",
            {
                "database_name": request.url_vars["database"],
                "table_name": request.url_vars["table"],
                "actions": [
                    "insert-row",
                    "delete-row",
                    "update-row",
                    "alter-table",
                    "drop-table",
                ],
                "groups": groups,
                "group_sizes": group_sizes,
                "group_permissions": current_group_permissions,
                "user_permissions": current_user_permissions,
                "audit_log": audit_log.rows,
                "valid_actor_ids": await get_acl_actor_ids(datasette),
            },
            request=request,
        )
    )



from datasette.plugins import pm
from datasette.utils import await_me_maybe


async def can_edit_permissions(datasette, actor):
    return await datasette.permission_allowed(actor, "datasette-acl")


def generate_changes_message(changes_made, noun):
    messages = []
    for action, changes in changes_made.items():
        for name, permission in changes:
            messages.append(f"{action}: {noun} '{name}' can {permission}")
    if not messages:
        return None
    message = ", ".join(messages)
    # Capitalize first letter
    return message[0].upper() + message[1:]


async def get_acl_actor_ids(datasette):
    actor_ids = []
    for hook in pm.hook.datasette_acl_actor_ids(datasette=datasette):
        actor_ids.extend(await await_me_maybe(hook))
    return actor_ids


async def validate_actor_id(datasette, actor_id):
    actor_ids = await get_acl_actor_ids(datasette)
    if not actor_ids:
        # No validation has been configured
        return True
    else:
        return actor_id in actor_ids



from datasette import hookimpl, Permission
from datasette.events import CreateTableEvent
from datasette.utils import actor_matches_allow
from datasette.plugins import pm
from datasette_acl.utils import can_edit_permissions
from datasette_acl.views.table_acls import manage_table_acls
from datasette_acl.views.groups import manage_groups, manage_group
from . import hookspecs
import json
import sys
import time

pm.add_hookspecs(hookspecs)

CREATE_TABLES_SQL = """
create table if not exists acl_resources (
    id integer primary key,
    database text not null,
    resource text,
    unique(database, resource)
);

create table if not exists acl_actions (
    id integer primary key,
    name text not null unique
);

-- new table for groups
create table if not exists acl_groups (
    id integer primary key,
    name text not null unique,
    deleted integer
);

-- new table for actor-group relationships
create table if not exists acl_actor_groups (
    actor_id text,
    group_id integer,
    primary key (actor_id, group_id),
    foreign key (group_id) references acl_groups(id)
);

-- Group membership audit log
create table if not exists acl_groups_audit (
    id integer primary key,
    timestamp text default (datetime('now')),
    operation_by text,
    operation text check (operation in ('added', 'removed', 'created', 'deleted')),
    group_id integer,
    actor_id text,
    foreign key (group_id) references acl_groups(id)
);

create table if not exists acl (
    acl_id integer primary key,
    actor_id text,
    group_id integer,
    resource_id integer,
    action_id integer,
    foreign key (group_id) references acl_groups(id),
    foreign key (resource_id) references acl_resources(id),
    foreign key (action_id) references acl_actions(id),
    check ((actor_id is null) != (group_id is null)),
    unique(actor_id, group_id, resource_id, action_id)
);

-- ACL audit log
create table if not exists acl_audit (
    id integer primary key,
    timestamp text default (datetime('now')),
    operation_by text,
    operation text check (operation in ('added', 'removed')),
    action_id integer,
    resource_id integer,
    group_id integer,
    actor_id text,
    foreign key (group_id) references acl_groups(id),
    foreign key (resource_id) references acl_resources(id),
    foreign key (action_id) references acl_actions(id)
)
"""

ACL_RESOURCE_PAIR_SQL = """
with actor_groups as (
  select group_id
  from acl_actor_groups
  where actor_id = :actor_id
),
target_resource as (
  select id
  from acl_resources
  where database = :database and resource = :resource
),
target_action as (
  select id
  from acl_actions
  where name = :action
),
combined_permissions as (
    select resource_id, action_id
    from acl
    where actor_id = :actor_id
  union
    select resource_id, action_id
    from acl
    where group_id in (select group_id from actor_groups)
)
select count(*)
  from combined_permissions
  where resource_id = (select id from target_resource)
  and action_id = (select id from target_action)
"""

EXPECTED_GROUPS_SQL = """
with expected_groups as (
  select value as group_name
  from json_each(:expected_groups_json)
),
dynamic_groups as (
  select value as group_name
  from json_each(:dynamic_groups)
),
actual_groups as (
  select g.name as group_name
  from acl_groups g
  join acl_actor_groups ug on g.id = ug.group_id
  where ug.actor_id = :actor_id
)
select
  'should-add' as status,
  eg.group_name
from expected_groups eg
where eg.group_name not in (select group_name from actual_groups)
  union all
select
  'should-remove' as status,
  ag.group_name
from actual_groups ag
where ag.group_name not in (select group_name from expected_groups)
and ag.group_name in (select group_name from dynamic_groups)
  union all
select
  'current' as status,
  group_name
from actual_groups
"""


@hookimpl
def startup(datasette):
    async def inner():
        db = datasette.get_internal_database()
        await db.execute_write_script(CREATE_TABLES_SQL)
        # Ensure permissions are in the DB
        await db.execute_write_many(
            """
            insert or ignore into acl_actions (name) values (:name)
        """,
            [{"name": n} for n in datasette.permissions.keys()],
        )
        # And any dynamic groups
        config = datasette.plugin_config("datasette-acl") or {}
        groups = config.get("dynamic-groups")
        if groups:
            await db.execute_write_many(
                "insert or ignore into acl_groups (name) values (:name)",
                [{"name": name} for name in groups.keys()],
            )

    return inner


class OneSecondCache:
    def __init__(self):
        self.cache = {}

    def set(self, key, value):
        self.cache[key] = (value, time.monotonic() + 1)

    def get(self, key):
        if key in self.cache:
            value, expiration_time = self.cache[key]
            if time.monotonic() < expiration_time:
                return value
            else:
                del self.cache[key]
        return None

    def clear_expired(self):
        current_time = time.monotonic()
        self.cache = {k: v for k, v in self.cache.items() if v[1] > current_time}


one_second_cache = OneSecondCache()


async def update_dynamic_groups(datasette, actor, skip_cache=False):
    if not actor or not actor.get("id"):
        return
    if (not skip_cache) and one_second_cache.get(actor["id"]):
        # Don't do this more than once a second per actor
        return
    one_second_cache.set(actor["id"], 1)
    config = datasette.plugin_config("datasette-acl") or {}
    groups = config.get("dynamic-groups")
    if not groups:
        return
    # Figure out the groups the user should be in
    should_have_groups = set(
        group_name
        for group_name, allow_block in groups.items()
        if actor_matches_allow(actor, allow_block)
    )
    db = datasette.get_internal_database()
    result = await db.execute(
        EXPECTED_GROUPS_SQL,
        {
            "actor_id": actor["id"],
            "expected_groups_json": json.dumps(list(should_have_groups)),
            "dynamic_groups": json.dumps(list(groups.keys())),
        },
    )
    should_add = []
    should_remove = []
    for row in result.rows:
        if row["status"] == "should-add":
            should_add.append(row["group_name"])
        elif row["status"] == "should-remove":
            should_remove.append(row["group_name"])
    # Add/remove groups as needed
    for group_name in should_add:
        # Make sure the group exists
        await db.execute_write(
            "insert or ignore into acl_groups (name) VALUES (:name);",
            {"name": group_name},
        )
        await db.execute_write(
            """
            insert into acl_actor_groups (
                actor_id, group_id
            ) values (
                :actor_id,
                (select id from acl_groups where name = :group_name)
            )""",
            {"actor_id": actor["id"], "group_name": group_name},
        )
        await db.execute_write(
            """
            insert into acl_groups_audit (
                operation_by, operation, group_id, actor_id
            ) values (
                null,
                'added',
                (select id from acl_groups where name = :group_name),
                :actor_id
            )
        """,
            {
                "group_name": group_name,
                "actor_id": actor["id"],
            },
        )
    for group_name in should_remove:
        await db.execute_write(
            """
            delete from acl_actor_groups
            where actor_id = :actor_id
            and group_id = (select id from acl_groups where name = :group_name)
            """,
            {"actor_id": actor["id"], "group_name": group_name},
        )
        await db.execute_write(
            """
            insert into acl_groups_audit (
                operation_by, operation, group_id, actor_id
            ) values (
                null,
                'removed',
                (select id from acl_groups where name = :group_name),
                :actor_id
            )
        """,
            {
                "group_name": group_name,
                "actor_id": actor["id"],
            },
        )


@hookimpl
def permission_allowed(datasette, actor, action, resource):
    if not resource or len(resource) != 2:
        return None

    async def inner():
        if not actor or not actor.get("id"):
            return None
        await update_dynamic_groups(
            datasette, actor, skip_cache=hasattr(sys, "_pytest_running")
        )
        db = datasette.get_internal_database()
        result = await db.execute(
            ACL_RESOURCE_PAIR_SQL,
            {
                "actor_id": actor["id"],
                "database": resource[0],
                "resource": resource[1],
                "action": action,
            },
        )
        return result.single_value() or None

    return inner


@hookimpl
def register_permissions(datasette):
    return [
        Permission(
            name="datasette-acl",
            abbr=None,
            description="Configure permissions",
            takes_database=False,
            takes_resource=False,
            default=False,
        )
    ]


@hookimpl
def table_actions(datasette, actor, database, table):
    async def inner():
        if await can_edit_permissions(datasette, actor):
            return [
                {
                    "href": datasette.urls.table(database, table) + "/-/acl",
                    "label": "Manage table permissions",
                    "description": "Control who can  write, and delete rows in this table",
                }
            ]

    return inner


@hookimpl
def track_event(datasette, event):
    async def inner():
        config = datasette.plugin_config("datasette-acl") or {}
        if not config.get("table-creator-permissions"):
            return
        if not isinstance(event, CreateTableEvent):
            return
        if not event.actor:
            return
        # Add ACLs for the user who created the table
        db = datasette.get_internal_database()
        # Ensure resource exists for table
        await db.execute_write(
            "INSERT OR IGNORE INTO acl_resources (database, resource) VALUES (?, ?);",
            [event.database, event.table],
        )
        resource_id = (
            await db.execute(
                "SELECT id FROM acl_resources WHERE database = ? AND resource = ?",
                [event.database, event.table],
            )
        ).single_value()
        await db.execute_write_many(
            """
            INSERT INTO acl (actor_id, group_id, resource_id, action_id)
            VALUES (
                :actor_id,
                null,
                :resource_id,
                (SELECT id FROM acl_actions WHERE name = :action_name)
            )
            """,
            [
                {
                    "actor_id": event.actor["id"],
                    "action_name": action_name,
                    "resource_id": resource_id,
                }
                for action_name in config["table-creator-permissions"]
            ],
        )

    return inner


@hookimpl
def menu_links(datasette, actor):
    async def inner():
        if await can_edit_permissions(datasette, actor):
            return [
                {
                    "href": datasette.urls.path("/-/acl/groups"),
                    "label": "Manage user groups",
                }
            ]

    return inner


@hookimpl
def register_routes():
    return [
        ("^/(?P<database>[^/]+)/(?P<table>[^/]+)/-/acl$", manage_table_acls),
        ("^/-/acl/groups$", manage_groups),
        ("^/-/acl/groups/(?P<name>[^/]+)$", manage_group),
    ]



from collections import namedtuple
import pytest

ManageGroupTest = namedtuple(
    "ManageGroupTest",
    (
        "description",
        "setup_post_data",
        "post_data",
        "expected_members",
        "expected_audit_rows",
    ),
)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ManageGroupTest._fields,
    (
        ManageGroupTest(
            description="Add user to group",
            setup_post_data={},
            post_data={"add": "terry"},
            expected_members={"terry"},
            expected_audit_rows=[
                {"operation_by": "root", "operation": "added", "actor_id": "terry"}
            ],
        ),
        ManageGroupTest(
            description="Remove user from group",
            setup_post_data={"add": "terry"},
            post_data={"remove": "terry"},
            expected_members=set(),
            expected_audit_rows=[
                {"operation_by": "root", "operation": "removed", "actor_id": "terry"},
                {"operation_by": "root", "operation": "added", "actor_id": "terry"},
            ],
        ),
    ),
)
async def test_manage_table_permissions(
    ds,
    csrftoken,
    description,
    setup_post_data,
    post_data,
    expected_members,
    expected_audit_rows,
):
    internal_db = ds.get_internal_database()

    if setup_post_data:
        setup_response = await ds.client.post(
            "/-/acl/groups/dev",
            data={**setup_post_data, "csrftoken": csrftoken},
            cookies={
                "ds_actor": ds.client.actor_cookie({"id": "root"}),
                "ds_csrftoken": csrftoken,
            },
        )
        assert setup_response.status_code == 302

    response = await ds.client.post(
        "/-/acl/groups/dev",
        data={**post_data, "csrftoken": csrftoken},
        cookies={
            "ds_actor": ds.client.actor_cookie({"id": "root"}),
            "ds_csrftoken": csrftoken,
        },
    )
    assert response.status_code == 302

    # Check group members
    group_members = await get_group_members(internal_db, "dev")
    assert group_members == expected_members

    # Check audit logs
    AUDIT_SQL = """
        select
          operation_by, operation, actor_id
        from acl_groups_audit
        where group_id = (select id from acl_groups where name = 'dev')
        order by id desc
    """
    audit_rows = [dict(r) for r in (await internal_db.execute(AUDIT_SQL))]
    assert audit_rows == expected_audit_rows


@pytest.mark.asyncio
async def test_cannot_edit_dynamic_group(ds, csrftoken):
    db = ds.get_internal_database()

    # Adding to dev should work, adding to staff should fail
    for group in ("staff", "dev"):
        await ds.client.post(
            f"/-/acl/groups/{group}",
            data={"add": "tony2", "csrftoken": csrftoken},
            cookies={
                "ds_actor": ds.client.actor_cookie({"id": "root"}),
                "ds_csrftoken": csrftoken,
            },
        )
    assert await get_group_members(db, "staff") == set()
    assert await get_group_members(db, "dev") == {"tony2"}


async def get_group_members(db, group):
    return {
        d["actor_id"]
        for d in (
            await db.execute(
                """
        select actor_id from acl_actor_groups
        where group_id = (select id from acl_groups where name = ?)
    """,
                [group],
            )
        ).rows
    }


@pytest.mark.asyncio
async def test_deleted_group(ds):
    db = ds.get_internal_database()
    await db.execute_write(
        "insert into acl_groups (name, deleted) values ('deleted', 1)"
    )
    list_response = await ds.client.get(
        "/-/acl/groups",
        cookies={
            "ds_actor": ds.client.actor_cookie({"id": "root"}),
        },
    )
    # Should link to staff but not to deleted
    assert "/-/acl/groups/staff" in list_response.text
    assert "/-/acl/groups/deleted" not in list_response.text

    # deleted page should still 200 but it should say it is deleted
    page_response = await ds.client.get(
        "/-/acl/groups/deleted",
        cookies={
            "ds_actor": ds.client.actor_cookie({"id": "root"}),
        },
    )
    assert page_response.status_code == 200
    assert "has been deleted" in page_response.text


@pytest.mark.asyncio
async def test_create_delete_group(ds, csrftoken):
    internal_db = ds.get_internal_database()

    # Create a group
    create_group_response = await ds.client.post(
        "/-/acl/groups",
        data={"new_group": "sales", "csrftoken": csrftoken},
        cookies={
            "ds_actor": ds.client.actor_cookie({"id": "root"}),
            "ds_csrftoken": csrftoken,
        },
    )
    assert create_group_response.status_code == 302
    assert create_group_response.headers["location"] == "/-/acl/groups/sales"

    async def get_members():
        return {
            r[0]
            for r in (
                await internal_db.execute(
                    """
            select actor_id
            from acl_actor_groups
            where group_id = (select id from acl_groups where name = 'sales')
        """
                )
            )
        }

    assert await get_members() == set()

    # Add sally, sam and paulo
    for actor_id in ("sally", "sam", "paulo"):
        add_response = await ds.client.post(
            f"/-/acl/groups/sales",
            data={"add": actor_id, "csrftoken": csrftoken},
            cookies={
                "ds_actor": ds.client.actor_cookie({"id": "root"}),
                "ds_csrftoken": csrftoken,
            },
        )
        assert add_response.status_code == 302
        assert add_response.headers["location"] == "/-/acl/groups/sales#focus-add"
    # Check the group has those members
    assert await get_members() == {"sally", "sam", "paulo"}
    # Deleting this group should first remove the members
    delete_group_response = await ds.client.post(
        "/-/acl/groups/sales",
        data={"delete_group": "1", "csrftoken": csrftoken},
        cookies={
            "ds_actor": ds.client.actor_cookie({"id": "root"}),
            "ds_csrftoken": csrftoken,
        },
    )
    assert delete_group_response.status_code == 302
    assert delete_group_response.headers["location"] == "/-/acl/groups"

    assert await get_members() == set()

    # Should be marked as deleted
    assert (
        await internal_db.execute("select deleted from acl_groups where name = 'sales'")
    ).single_value() == 1

    # Check the audit log
    audit_rows = [
        dict(r)
        for r in (
            await internal_db.execute(
                """
        select
          operation_by, operation, actor_id
        from acl_groups_audit
        where group_id = (select id from acl_groups where name = 'sales')
        order by id desc
    """
            )
        )
    ]
    assert audit_rows == [
        {"operation_by": "root", "operation": "deleted", "actor_id": None},
        {"operation_by": "root", "operation": "removed", "actor_id": "paulo"},
        {"operation_by": "root", "operation": "removed", "actor_id": "sam"},
        {"operation_by": "root", "operation": "removed", "actor_id": "sally"},
        {"operation_by": "root", "operation": "added", "actor_id": "paulo"},
        {"operation_by": "root", "operation": "added", "actor_id": "sam"},
        {"operation_by": "root", "operation": "added", "actor_id": "sally"},
        {"operation_by": "root", "operation": "created", "actor_id": None},
    ]



from datasette import hookimpl
from datasette_acl.utils import get_acl_actor_ids
from datasette.plugins import pm
import pytest


@pytest.fixture
def register_plugin():
    class TestActorIdsPlugin:
        __name__ = "TestActorIdsPlugin"

        @hookimpl
        def datasette_acl_actor_ids(self, datasette):
            async def inner():
                db = datasette.get_internal_database()
                return [
                    r[0] for r in (await db.execute("select username from users")).rows
                ]

            return inner

    pm.register(TestActorIdsPlugin(), name="undo")
    try:
        yield
    finally:
        pm.unregister(name="undo")


@pytest.mark.asyncio
async def test_datasette_acl_actor_ids_hook(ds, csrftoken, register_plugin):
    plugins_response = await ds.client.get("/-/plugins.json")
    assert any(
        plugin
        for plugin in plugins_response.json()
        if plugin["name"] == "TestActorIdsPlugin"
    )
    await ds.get_internal_database().execute_write_script(
        """
        create table if not exists users (username text primary key);
        insert or ignore into users (username) values ('one');
        insert or ignore into users (username) values ('two');
        """
    )
    actor_ids = await get_acl_actor_ids(ds)
    assert actor_ids == ["one", "two"]

    to_test = (
        ("one", True),
        ("three", False),
    )

    # Check these are validated when editing permissions
    for actor_id, should_work in to_test:
        response = await ds.client.post(
            "/db/t/-/acl",
            data={
                "new_actor_id": actor_id,
                "new_user_insert-row": "on",
                "csrftoken": csrftoken,
            },
            cookies={
                "ds_actor": ds.client.actor_cookie({"id": "root"}),
                "ds_csrftoken": csrftoken,
            },
        )
        assert response.status_code == 302
        messages = ds.unsign(response.cookies["ds_messages"], "messages")
        if should_work:
            # Should be a positive message
            assert messages[0][1] == ds.INFO
        else:
            assert messages[0] == ["That user ID is not valid", ds.ERROR]

    # And when editing group members
    for actor_id, should_work in to_test:
        response = await ds.client.post(
            "/-/acl/groups/dev",
            data={
                "add": actor_id,
                "csrftoken": csrftoken,
            },
            cookies={
                "ds_actor": ds.client.actor_cookie({"id": "root"}),
                "ds_csrftoken": csrftoken,
            },
        )
        assert response.status_code == 302
        messages = ds.unsign(response.cookies["ds_messages"], "messages")
        if should_work:
            # Should be a positive message
            assert messages[0][1] == ds.INFO
        else:
            assert messages[0] == ["That user ID is not valid", ds.ERROR]



from pluggy import HookspecMarker

hookspec = HookspecMarker("datasette")


@hookspec
def datasette_acl_actor_ids(datasette):
    """
    List of actor IDs that can be autocompleted against when editing permissions
    """



from datasette.app import Datasette
import pytest


@pytest.mark.asyncio
async def test_can_startup_with_no_configuration():
    datasette = Datasette()
    await datasette.invoke_startup()
    assert (await datasette.client.get("/")).status_code == 200
